/*
 * Copyright (c) 1999  Albert Dorofeev <Albert@mail.dma.be>
 * For the updates see http://bewoner.dma.be/Albert/
 *
 * This software is distributed under GPL. For details see LICENSE file.
 */

#ifndef _read_mem_h_
#define _read_mem_h_

void open_meminfo();
void read_meminfo();
void close_meminfo();

#endif /* _read_mem_h_ */

